#include "fintek_api.h"
#include <pthread.h>

#define FINTEK_GPIO_INT_TRIGGER_PATH	"/sys/bus/platform/devices/fintek-gpio-int/irq-%d"
#define FORCE_GROUP_MODE	0
#define MAX_SET			4
#define NEED_LOCK		1

static pthread_attr_t pthread_attr;
static int m_fd[MAX_SET];
static pthread_t m_tid[MAX_SET];
static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

unsigned int m_idx[MAX_SET];
unsigned int m_irq[MAX_SET];

#define SECOND (1000000LL)
#define MSECOND (1000LL)

static void time_start(struct timeval *tv)
{
	gettimeofday(tv, NULL);
}

static time_t time_delta(struct timeval *start, struct timeval *stop)
{
	time_t secs, usecs;

	secs = stop->tv_sec - start->tv_sec;
	usecs = stop->tv_usec - start->tv_usec;
	if (usecs < 0) {
		secs--;
		usecs += SECOND;
	}

	return secs * SECOND + usecs;
}

static time_t time_stop(struct timeval *begin)
{
	struct timeval end;
	gettimeofday(&end, NULL);
	return time_delta(begin, &end);
}

void usage(int argc, char *argv[])
{
	fprintf(stderr, "\n\n\n\nFintek GPIO interrupt Demo:\n");
	fprintf(stderr, "%s <pin,irq> ...\n\n", argv[0]);
	fprintf(stderr, "Param List:\n");
	fprintf(stderr, "\tpin: number of GPIO idx, like GPIO12, idx=0x12\n");
	fprintf(stderr, "\tirq: irq channel\n\n");
	fprintf(stderr, "\texample: %s 0x80,5\n", argv[0]);
}

void lock()
{
#if NEED_LOCK
	pthread_mutex_lock(&mutex);
#endif
}

void unlock()
{
#if NEED_LOCK
	pthread_mutex_unlock(&mutex);
#endif
}

void *do_waiting_int(void *arg_irq)
{
	int irq = *((int *)arg_irq);
	int fd;
	int cnt = 0;
	int size;
	unsigned int data;
	char ch;
	int i;
	int idx = -1;
	unsigned char buf[512];
	struct timeval start, end;

	for (i = 0; i < MAX_SET; ++i) {
		if (m_irq[i] == irq) {
			idx = m_idx[i];
			break;
		}
	}

	if (idx == -1) {
		printf("%s: can't found irq: %d pair\n", __func__, irq);
		return NULL;
	}

	while (1) {
		sprintf(buf, FINTEK_GPIO_INT_TRIGGER_PATH, irq);
		fd = open(buf, O_RDONLY);
		if (fd < 0) {
			printf("open failed\n");
			return 0;
		}

		while (1) {
			size = read(fd, &ch, 1);
			if (size < 0) {
				perror("read");
				break;
			}

			size = lseek(fd, SEEK_SET, 0);
			if (size < 0) {
				perror("lseek");
				break;
			}

			lock();
			printf("in\n");
			while (1) {
				CHECK_RET_TYPE(_GetGpioIntStatus(idx >> 4, &data), void *);
				CHECK_RET_TYPE(_ClearGpioIntStatus(idx >> 4, data), void *);
				if ((data & BIT(idx & 0x0f)) == 0) {
					printf("out\n");
					break;
				}

				CHECK_RET_TYPE(_GetGpioInputDataIdx(idx, &data), void *);
				printf("GPIO%02x: cnt: %d, data: %d\n", idx, cnt++, data);
			}
			unlock();
		}

		close(fd);

	}

	return NULL;
}

int gpio_int_demo(int argc, char *argv[])
{
	unsigned int data = 0;
	int status;
	int i;

	if (argc < 2 || argc > 5) {
		usage(argc, argv);
		exit(3);
	}

	for (i = 0; i < MAX_SET && i < argc - 1; ++i) {
		sscanf(argv[i + 1], "%x,%d", &m_idx[i], &m_irq[i]);
		if (!m_irq[i]) {
			printf("irq can't be 0\n");
			usage(argc, argv);
			exit(3);
		}
		//printf("get m_idx: %x %d\n", m_idx[i], m_irq[i]);
	}

	for (i = 0; i < MAX_SET && i < argc - 1; ++i) {
		CHECK_RET(_EnableGPIO(m_idx[i], eGPIO_Mode_Enable));
		printf("idx: 0x%x => set:%x, count:%x, irq: %x\n", m_idx[i], (m_idx[i] >> 4) & 0xf,
		       (m_idx[i] >> 0) & 0xf, m_irq[i]);

		CHECK_RET(_SetGpioOutputEnableIdx(m_idx[i], eGPIO_Direction_In));
		printf("Dir: 0x%x => %s\n", data, "IN");

		// pull high if available (not error report)
		_SetGpioPullMode(m_idx[i], eGPIO_Pull_High);

		CHECK_RET(_DisableGpioInt(m_idx[i] >> 4, BIT(m_idx[i] & 0x0f)));
		CHECK_RET(_ClearGpioIntStatus(m_idx[i] >> 4, BIT(m_idx[i] & 0x0f)));
		CHECK_RET(_EnableGpioInt(m_idx[i] >> 4, BIT(m_idx[i] & 0x0f), m_irq[i]));
	}

	// do start pthreads
	pthread_attr_init(&pthread_attr);

	for (i = 0; i < MAX_SET && i < argc - 1; ++i)
		pthread_create(&m_tid[i], &pthread_attr, do_waiting_int, &m_irq[i]);

	// waiting end
	for (i = 0; i < MAX_SET && i < argc - 1; ++i)
		pthread_join(m_tid[i], NULL);

	fflush(stdout);

	return 0;
}

int main(int argc, char *argv[])
{
	int nRet = 0;
	sFintek_sio_data sio_data;

	/* Please set this to 1 enable more debug message */
	set_debug(0);

	fprintf(stderr, "FINTEK Lib Version: %s\n", getFintekLibVersion());

	if (nRet = init_fintek_sio(eSIO_TYPE_F81866, 0, &sio_data)) {
		fprintf(stderr, "init_fintek_sio error\n");
		exit(3);
	}

	ActiveSIO(sio_data.ic_port, sio_data.key);

	/* GPIO interrupt test, need driver */
	gpio_int_demo(argc, argv);

	DeactiveSIO(sio_data.ic_port);

	return 0;
}
