/*
 *  Probe for F81216A LPC to 4 UART
 *
 *  Based on drivers/tty/serial/8250_pnp.c, by Russell King, et al
 *
 *  Copyright (C) 2014 Ricardo Ribalda, Qtechnology A/S
 *
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 */
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/pnp.h>
#include <linux/kernel.h>
#include <linux/serial_core.h>
#include <linux/serial_8250.h>
#include <linux/serial_reg.h>
#include <linux/version.h>

#define DRVNAME	"fintek-gpio-int"
#define DRVVER	"v1.1"

#define TARGET_PIN BIT(6)

#define LDN 0x07
#define EXIT_KEY 0xAA
#define ADDR_PORT 0
#define DATA_PORT 1
#define	MAXDEV 4

static unsigned int port = 0x2e;
module_param(port, uint, S_IRUGO);
MODULE_PARM_DESC(port, "IO port, default = 0x2e");

static unsigned int key = 0x87;
module_param(key, uint, S_IRUGO);
MODULE_PARM_DESC(key, "Activate key, default = 0x87");

static unsigned int ldn = 0x06;
module_param(ldn, uint, S_IRUGO);
MODULE_PARM_DESC(ldn, "Activate key, default = 0x06");

#if 0
static unsigned int irq = 5;
module_param(irq, uint, S_IRUGO);
MODULE_PARM_DESC(irq, "IRQ, default = 5");
#else
static int irq[MAXDEV];
module_param_array(irq, int, NULL, 0444);
MODULE_PARM_DESC(irq, "IRQ number, max 4 entry");
#endif

struct fintek_sio {
	u16 pid;
	u16 base_port;
	u8 key;
	u8 triggered[MAXDEV];
	spinlock_t lock;
	struct completion irq_comp[MAXDEV];
	struct device_attribute dev_irq[MAXDEV];
};

static struct fintek_sio m_sio_data;

static u8 sio_read_reg(struct fintek_sio *pdata, u8 reg)
{
	outb(reg, pdata->base_port + ADDR_PORT);
	return inb(pdata->base_port + DATA_PORT);
}

#if 0
static void sio_write_reg(struct fintek_sio *pdata, u8 reg, u8 data)
{
	outb(reg, pdata->base_port + ADDR_PORT);
	outb(data, pdata->base_port + DATA_PORT);
}

static void sio_write_mask_reg(struct fintek_sio *pdata, u8 reg, u8 mask,
			       u8 data)
{
	u8 tmp;

	tmp = (sio_read_reg(pdata, reg) & ~mask) | (mask & data);
	sio_write_reg(pdata, reg, tmp);
}
#endif

static int fintek_sio_enter_key(u16 base_port, u8 key)
{
	if (!request_muxed_region(base_port, 2, DRVNAME))
		return -EBUSY;

	outb(key, base_port + ADDR_PORT);
	outb(key, base_port + ADDR_PORT);
	return 0;
}

static void fintek_8250_exit_key(u16 base_port)
{

	outb(EXIT_KEY, base_port + ADDR_PORT);
	release_region(base_port + ADDR_PORT, 2);
}

static irqreturn_t sio_interrupt(int local_irq, void *dev_id)
{
	struct fintek_sio *sio = dev_get_platdata(dev_id);
	unsigned long flags;
	int i;

	for (i = 0; i < MAXDEV; ++i) {
		if (local_irq == irq[i])
			break;
	}

	if (i >= MAXDEV) {
		pr_warn("%s: can't found match irq handler\n", __func__);
		return IRQ_NONE;
	}

	spin_lock_irqsave(&sio->lock, flags);
	sio->triggered[i] = 1;
	complete(&sio->irq_comp[i]);
	spin_unlock_irqrestore(&sio->lock, flags);

	return IRQ_HANDLED;
}

#if 0
static int configure_gpio_with_read_interrupt(struct fintek_sio *sio_data)
{
	int status = 0;

	fintek_sio_enter_key(sio_data->base_port, sio_data->key);

	sio_write_reg(sio_data, LDN, ldn); /* Set LDN to 6*/

	sio_write_reg(sio_data, 0x8e, 0x0); /* GPIO8x Disable all Interrupt */

	sio_write_reg(sio_data, 0x73, 0x00); /* GPIO8x IRQ set to 0 */
	sio_write_reg(sio_data, 0x7e, 0x08); /* GPIO8x share IRQ */
	sio_write_reg(sio_data, 0x7f, 0x40); /* GPIO8x IRQ High/Edge Mode */

	sio_write_mask_reg(sio_data, 0x88, TARGET_PIN, 0); /* GPIO8x  input mode */

	/*
	 * Beware, due to request_muxed_region() in handler, we must ensure
	 * using THREAD_IRQ
	 */
	status = request_irq(irq, sio_interrupt, 0, DRVNAME,
				sio_data);
	if (status) {
		printk(KERN_INFO "%s: request_irq failed: %02x\n", __func__,
				status);
		goto end;
	}

	sio_write_reg(sio_data, 0x8f, TARGET_PIN); /* GPIO8x Clear SMI status */
	sio_write_reg(sio_data, 0x73, irq); /* GPIO8x IRQ set to IRQ */
	sio_write_reg(sio_data, 0x8e, TARGET_PIN); /* GPIO8x Enable Interrupt */

	fintek_8250_exit_key(sio_data->base_port);

end:
	return status;
}

static int set_gpio_fast(struct fintek_sio *sio_data)
{
	u8 data = 0;
	int count = 1000;

	fintek_sio_enter_key(sio_data->base_port, sio_data->key);

	sio_write_reg(sio_data, LDN, ldn); /* Set LDN to 6*/
	sio_write_mask_reg(sio_data, 0x88, TARGET_PIN, TARGET_PIN); /* GPIO8x  output mode */
	sio_write_mask_reg(sio_data, 0x8b, TARGET_PIN, TARGET_PIN); /* GPIO8x  output push-pull mode */

	outb(0x89, sio_data->base_port + ADDR_PORT);

	while (count--) {
		data = data ? 0 : TARGET_PIN;
		//sio_write_mask_reg(sio_data, 0x89, TARGET_PIN, data);
		//sio_write_reg(sio_data, 0x89, data);
		outb(data, sio_data->base_port + DATA_PORT);
	}

	fintek_8250_exit_key(sio_data->base_port);
	return 0;
}
#endif

static int check_ic(struct fintek_sio *sio_data)
{
	int status = 0;
	int vid;

	status = fintek_sio_enter_key(sio_data->base_port, sio_data->key);
	if (status)
		return status;

	vid = sio_read_reg(sio_data, 0x23) << 8;
	vid |= sio_read_reg(sio_data, 0x24) << 0;

	if (vid != 0x1934)
		status = -ENODEV;

	fintek_8250_exit_key(sio_data->base_port);

	return status;
}

static ssize_t gpio_show(struct device *dev, struct device_attribute *attr,
				char *buf)
{
	struct fintek_sio *sio = dev_get_platdata(dev);
	unsigned long flags;
	long status;
	int i;
	int found = 0;

	for (i = 0; i < MAXDEV; ++i) {
		if (!irq[i])
			break;

		if (!strcmp(sio->dev_irq[i].attr.name, attr->attr.name)) {
			found = 1;
			break;
		}
	}

	if (!found) {
		pr_warn("%s: can't found match dev_irq index\n", __func__);
		return -EINVAL;
	}

	while (1) {
		status = wait_for_completion_interruptible_timeout(&sio->irq_comp[i],
			msecs_to_jiffies(10));
		if (status < 0)
			return -EINTR;

		spin_lock_irqsave(&sio->lock, flags);
		if (sio->triggered[i]) {
			sio->triggered[i] = 0;
			spin_unlock_irqrestore(&sio->lock, flags);
			break;
		}
		spin_unlock_irqrestore(&sio->lock, flags);
	}

	return sprintf(buf, "1");
}

static struct platform_device *gpio_dev;
//static DEVICE_ATTR_RO(gpio);

static int __init fintek_gpio_device_add(const struct fintek_sio *sio)
{
	int err;

	gpio_dev = platform_device_alloc(DRVNAME, -1);
	if (!gpio_dev)
		return -ENOMEM;

	err = platform_device_add_data(gpio_dev, sio, sizeof(*sio));
	if (err) {
		pr_err(DRVNAME "Platform data allocation failed\n");
		goto err;
	}

	err = platform_device_add(gpio_dev);
	if (err) {
		pr_err(DRVNAME "Device addition failed\n");
		goto err;
	}

	return 0;

err:
	platform_device_put(gpio_dev);
	return err;
}

static int fintek_gpio_probe(struct platform_device *pdev)
{
	struct fintek_sio *sio = dev_get_platdata(&pdev->dev);
	char *name;
	int status;
	int i;

	for (i = 0; i < MAXDEV; ++i) {
		if (!irq[i])
			break;

		init_completion(&sio->irq_comp[i]);
		sio->triggered[i] = 0;

		name = devm_kzalloc(&pdev->dev, 8, GFP_KERNEL);
		if (!name)
			return -ENOMEM;

		sprintf(name, "irq-%d", irq[i]);

		sio->dev_irq[i].show = gpio_show;
		sio->dev_irq[i].attr.mode = 0444;
		sio->dev_irq[i].attr.name = name;

		status = device_create_file(&pdev->dev, &sio->dev_irq[i]);
		if (status)
			return -EPERM;

		status = request_irq(irq[i], sio_interrupt, IRQF_ONESHOT,
					DRVNAME, &pdev->dev);
		if (status) {
			device_remove_file(&pdev->dev, &sio->dev_irq[i]);

			printk(KERN_INFO "%s: request_irq failed: %02x\n",
					__func__, status);
			return status;
		}
	}

	return 0;
}

int fintek_gpio_remove(struct platform_device *pdev)
{
	struct fintek_sio *sio = dev_get_platdata(&pdev->dev);
	int i;

	for (i = 0; i < MAXDEV; ++i) {
		if (!irq[i])
			break;

		complete_all(&sio->irq_comp[i]);
		device_remove_file(&pdev->dev, &sio->dev_irq[i]);
		free_irq(irq[i], &pdev->dev);
	}

	return 0;
}

static struct platform_driver fintek_gpio_driver = {
	.driver = {
		.name	= DRVNAME,
	},

	.probe		= fintek_gpio_probe,
	.remove		= fintek_gpio_remove,
};

static int __init fintek_gpio_init(void)
{
	int status;
	int i;
#if 0
	m_sio_data.base_port = port;
	m_sio_data.key = key;

	pr_info("Fintek GPIO interrupt handler\n");
	pr_info("	port : %x\n", port);
	pr_info("	key  : %x\n", key);
	pr_info("	ldn  : %x\n", ldn);
	pr_info("	irq  : %x\n", irq);

	status = check_ic(&m_sio_data);
	if (status) {
		pr_info("cant found Fintek Super IO\n");
		return -ENODEV;
	}

	//configure_gpio_with_read_interrupt(&m_sio_data);

	status = request_irq(irq, sio_interrupt, IRQF_ONESHOT, DRVNAME,
				&m_sio_data);
	if (status) {
		printk(KERN_INFO "%s: request_irq failed: %02x\n", __func__,
				status);
		return status;
	}

#if 0
	status = device_create_file(&port->dev, &dev_attr_gpio);
	if (status) {
		free_irq(irq, &m_sio_data);
		return -EPERM;
	}
#endif
#else

	pr_info("Fintek GPIO interrupt handler "DRVVER"\n");
	pr_info("port : %x\n", port);
	pr_info("key  : %x\n", key);
	pr_info("ldn  : %x\n", ldn);

	for (i = 0; i < MAXDEV; ++i) {
		if (!irq[i])
			break;

		pr_info("handled irq : %x\n", irq[i]);
	}

	m_sio_data.base_port = port;
	m_sio_data.key = key;
	spin_lock_init(&m_sio_data.lock);

	status = check_ic(&m_sio_data);
	if (status) {
		pr_info("%s: cant found Fintek Super IO\n", __func__);
		return -ENODEV;
	}

	for (i = 0; i < MAXDEV; ++i) {
		if (!irq[i])
			break;
	}

	if (!i) {
		pr_warn("%s: must specify least 1 IRQ\n", __func__);
		return -EINVAL;
	}

	status = platform_driver_register(&fintek_gpio_driver);
	if (status)
		return status;

	status = fintek_gpio_device_add(&m_sio_data);
	if (status) {
		platform_driver_unregister(&fintek_gpio_driver);
		return status;
	}

#endif
	return 0;
}

static void __exit fintek_gpio_exit(void)
{
	//device_remove_file(&port->dev, &dev_attr_gpio);

	//free_irq(irq, &m_sio_data);

	platform_device_unregister(gpio_dev);
	platform_driver_unregister(&fintek_gpio_driver);
}

module_init(fintek_gpio_init);
module_exit(fintek_gpio_exit);

MODULE_LICENSE("GPL");
